@@ -1,5 +1,7 @@
+import logging
 import time
 from datetime import datetime
+from pathlib import Path
 from typing import Any, Literal, Self, TypeAlias
 
 import stjames
@@ -10,6 +12,9 @@ from stjames.optimization.freezing_string_method import FSMSettings
 from .protein import Protein
 from .utils import api_client
 
+logger = logging.getLogger(__name__)
+logger.setLevel(logging.INFO)
+
 RdkitMol: TypeAlias = Chem.rdchem.Mol | Chem.rdchem.RWMol
 StJamesMolecule: TypeAlias = stjames.Molecule
 
@@ -216,6 +221,54 @@ class Workflow(BaseModel):
             response = client.delete(f"/workflow/{self.uuid}/delete_workflow_data")
             response.raise_for_status()
 
+    def download_msa_files(self, msa_format: stjames.MSAFormat, path: Path | None = None) -> None:
+        if self.workflow_type != "msa":
+            raise ValueError("This workflow is not an MSA workflow.")
+
+        if path is None:
+            path = Path.cwd()
+
+        if not path.exists():
+            path.mkdir(parents=True, exist_ok=True)
+
+        with api_client() as client:
+            response = client.get(
+                f"/workflow/{self.uuid}/get_msa_files", params={"msa_format": msa_format.value}
+            )
+            response.raise_for_status()
+
+        with open(path / f"{self.name}-msa.tar.gz", "wb") as f:
+            f.write(response.content)
+
+    def download_dcd_files(self, replicates: list[int],
+                           name: str | None = None, path: Path | None = None) -> None:
+        """
+        Downloads DCD trajectory files for specified replicates
+
+        :param replicates: List of replicate indices to download
+        :param name: Optional custom name for the tar.gz file (defaults to workflow name)
+        :param path: Directory to save the file to (defaults to current directory)
+        :raises ValueError: if workflow is not a pose analysis MD workflow
+        :raises requests.HTTPError: if the request to the API fails
+        """
+        if self.workflow_type != "pose_analysis_md":
+            raise ValueError("This workflow is not a pose analysis molecular dynamics workflow.")
+
+        if path is None:
+            path = Path.cwd()
+
+        path.mkdir(parents=True, exist_ok=True)
+
+        with api_client() as client:
+            response = client.post(
+                f"/trajectory/{self.uuid}/trajectory_dcds", json=replicates
+            )
+            response.raise_for_status()
+
+        file_path = path / f"{name or self.name}.tar.gz"
+        with open(file_path, "wb") as f:
+            f.write(response.content)
+
 
 def submit_workflow(
     workflow_type: stjames.WORKFLOW_NAME,
@@ -272,6 +325,75 @@ def submit_workflow(
         return Workflow(**response.json())
 
 
+def batch_submit_workflow(
+    workflow_type: stjames.WORKFLOW_NAME,
+    workflow_data: dict[str, Any] | None = None,
+    initial_molecules: list[dict[str, Any] | StJamesMolecule | RdkitMol] | None = None,
+    initial_smileses: list[str] | None = None,
+    names: list[str] | None = None,
+    folder_uuid: str | None = None,
+    max_credits: int | None = None,
+) -> list[Workflow]:
+    """
+    Submits a batch of workflows to the API. Each workflow will be submitted with the
+    same workflow type, workflow data, and folder UUID, but with different initial molecules and/or
+    SMILES strings.
+
+    :param workflow_type: The type of workflow to submit.
+    :param workflow_data: A dictionary containing the data required to run the workflow.
+    :param initial_molecules: A list of molecule objects to use as the initial molecules in the
+    workflows.
+    At least one of a molecule or SMILES must be provided.
+    :param initial_smileses: A list of SMILES strings to use as the initial molecules in the
+    workflows.
+    At least one of a molecule or SMILES must be provided.
+    :param names: A list of names to give to the workflows.
+    :param folder_uuid: The UUID of the folder to store the workflow in.
+    :param max_credits: The maximum number of credits to use for the workflow. This applies per
+    workflow, not for the batch.
+    :return: A list of Workflow objects representing the submitted workflows.
+    :raises ValueError: If neither `initial_smiles` nor a valid `initial_molecule` is provided.
+    :raises HTTPError: If the API request fails.
+    """
+    if workflow_type not in stjames.WORKFLOW_MAPPING:
+        raise ValueError(
+            "Invalid workflow type. Must be one of:\n    " + "\n    ".join(stjames.WORKFLOW_MAPPING)
+        )
+
+    if initial_molecules and initial_smileses:
+        raise ValueError(
+            "You must provide either `initial_molecules` or `initial_smileses`, but not both."
+        )
+
+    if names:
+        if initial_molecules and len(names) != len(initial_molecules):
+            logger.warning(
+                "The number of names is not the same as the number of initial "
+                "molecules. Generic names of the form 'Batch Workflow {i}' "
+                "will be used."
+            )
+        if initial_smileses and len(names) != len(initial_smileses):
+            logger.warning(
+                "The number of names is not the same as the number of initial SMILES."
+                "Generic names of the form 'Batch Workflow {i}' will be used."
+            )
+
+    data = {
+        "names": names,
+        "folder_uuid": folder_uuid,
+        "workflow_type": workflow_type,
+        "workflow_data": workflow_data,
+        "max_credits": max_credits,
+        "initial_molecules": initial_molecules,
+        "initial_smileses": initial_smileses,
+    }
+
+    with api_client() as client:
+        response = client.post("/workflow/batch_submit", json=data)
+        response.raise_for_status()
+        return [Workflow(**workflow_data) for workflow_data in response.json()]
+
+
 def retrieve_workflow(uuid: str) -> Workflow:
     """
     Retrieves a workflow from the API.
@@ -402,7 +524,7 @@ def submit_basic_calculation_workflow(
             tasks=tasks,
             mode=mode,
         ),
-        engine=engine,
+        engine=engine or method.default_engine(),
     )
 
     data = {
@@ -551,8 +673,10 @@ def submit_solubility_workflow(
 
 
 def submit_pka_workflow(
-    initial_molecule: dict[str, Any] | StJamesMolecule | RdkitMol,
+    initial_molecule: dict[str, Any] | StJamesMolecule | RdkitMol | str,
     pka_range: tuple[int, int] = (2, 12),
+    method: Literal["aimnet2_wagen2024", "chemprop_nevolianis2025"] = "aimnet2_wagen2024",
+    solvent: str = "water",
     deprotonate_elements: list[int] | None = None,
     protonate_elements: list[int] | None = None,
     mode: str = "careful",
@@ -564,7 +688,10 @@ def submit_pka_workflow(
     Submits a pKa workflow to the API.
 
     :param initial_molecule: The molecule to calculate the pKa of.
+        Valid input types include `stjames.Molecule`, RDKit molecule, and SMILES.
     :param pka_range: The range of pKa values to calculate.
+    :param method: The algorithm used to compute pKa values.
+    :param solvent: The solvent in which pKa values will be computed.
     :param deprotonate_elements: The elements to deprotonate. Given by atomic number.
     :param protonate_elements: The elements to protonate. Given by atomic number.
     :param mode: The mode to run the calculation in. See
@@ -576,20 +703,30 @@ def submit_pka_workflow(
     :return: A Workflow object representing the submitted workflow.
     :raises requests.HTTPError: if the request to the API fails.
     """
-    if isinstance(initial_molecule, StJamesMolecule):
-        initial_molecule = initial_molecule.model_dump()
+    initial_smiles: str = ""
+    initial_stjames_mol: StJamesMolecule | None = None
+
+    if isinstance(initial_molecule, dict):
+        initial_stjames_mol = StJamesMolecule.model_validate(initial_molecule)
+    elif isinstance(initial_molecule, StJamesMolecule):
+        initial_stjames_mol = initial_molecule
     elif isinstance(initial_molecule, RdkitMol):
-        initial_molecule = StJamesMolecule.from_rdkit(initial_molecule, cid=0)
+        initial_stjames_mol = StJamesMolecule.from_rdkit(initial_molecule, cid=0)
+    elif isinstance(initial_molecule, str):
+        initial_smiles = initial_molecule
 
     protonate_elements = protonate_elements or [7]
     deprotonate_elements = deprotonate_elements or [7, 8, 16]
 
     workflow = stjames.pKaWorkflow(
-        initial_molecule=initial_molecule,
+        initial_molecule=initial_stjames_mol,
+        initial_smiles=initial_smiles,
         pka_range=pka_range,
         deprotonate_elements=deprotonate_elements,
         protonate_elements=protonate_elements,
         mode=mode,
+        solvent=solvent,
+        microscopic_pka_method=method,
     )
 
     data = {
@@ -597,7 +734,8 @@ def submit_pka_workflow(
         "folder_uuid": folder_uuid,
         "workflow_type": "pka",
         "workflow_data": workflow.model_dump(),
-        "initial_molecule": initial_molecule,
+        "initial_molecule": initial_stjames_mol.model_dump() if initial_stjames_mol else None,
+        "initial_smiles": initial_smiles,
         "max_credits": max_credits,
     }
 
@@ -1040,6 +1178,9 @@ def submit_docking_workflow(
     protein: str | Protein,
     pocket: list[list[float]],
     initial_molecule: dict[str, Any] | StJamesMolecule | RdkitMol | None = None,
+    executable: str = "vina",
+    scoring_function: str = "vinardo",
+    exhaustiveness: float = 8,
     do_csearch: bool = False,
     do_optimization: bool = False,
     do_pose_refinement: bool = False,
@@ -1050,15 +1191,18 @@ def submit_docking_workflow(
     """
     Submits a Docking workflow to the API.
 
-    :param protein: The protein to dock. Can be fed as a uuid or a Protein object.
+    :param protein: The protein to dock. Can be input as a uuid or a Protein object.
     :param initial_molecule: The initial molecule to be docked
+    :param executable: Which docking implementation to use.
+    :param scoring_function: Which docking scoring function to use.
+    :param exhaustiveness: Which exhaustiveness to employ.
     :param do_csearch: Whether to perform a conformational search on the ligand.
     :param do_optimization: Whether to perform an optimization on the ligand.
     :param do_pose_refinement: Whether or not to optimize output poses.
     :param name: The name of the workflow.
     :param folder_uuid: The UUID of the folder to place the workflow in.
     :param max_credits: The maximum number of credits to use for the workflow.
-    :return: A Workflow object representing the submitted IRC workflow.
+    :return: A Workflow object representing the submitted docking workflow.
     :raises requests.HTTPError: if the request to the API fails.
     """
 
@@ -1070,6 +1214,12 @@ def submit_docking_workflow(
     if isinstance(protein, Protein):
         protein = protein.uuid
 
+    docking_settings = {
+        "executable": executable,
+        "exhaustiveness": exhaustiveness,
+        "scoring_function": scoring_function,
+    }
+
     workflow = stjames.DockingWorkflow(
         initial_molecule=initial_molecule,
         target_uuid=protein,
@@ -1077,6 +1227,7 @@ def submit_docking_workflow(
         do_csearch=do_csearch,
         do_optimization=do_optimization,
         do_pose_refinement=do_pose_refinement,
+        docking_settings=docking_settings,
     )
 
     data = {
@@ -1283,3 +1434,146 @@ def submit_double_ended_ts_search_workflow(
         response = client.post("/workflow", json=data)
         response.raise_for_status()
         return Workflow(**response.json())
+
+
+def submit_pose_analysis_md_workflow(
+    protein: str | Protein,
+    initial_smiles: str,
+    num_trajectories: int = 1,
+    simulation_time_ns: int = 10,
+    ligand_residue_name: str = "LIG",
+    name: str = "Pose-Analysis MD Workflow",
+    folder_uuid: str | None = None,
+    max_credits: int | None = None,
+) -> Workflow:
+    """
+    Submits a pose-analysis MD workflow to the API.
+
+    :param protein: The *holo* protein on which MD will be run.
+        Can be input as a uuid or a Protein object.
+    :param initial_smiles: The SMILES for the ligand.
+    :param num_trajectories: The number of trajectories to run.
+    :param simulation_time_ns: How long to run the simulation for, in nanoseconds.
+    :param ligand_residue_name: The name of the residue corresponding to the ligand.
+    :param name: The name of the workflow.
+    :param folder_uuid: The UUID of the folder to place the workflow in.
+    :param max_credits: The maximum number of credits to use for the workflow.
+    :return: A Workflow object representing the submitted workflow.
+    :raises requests.HTTPError: if the request to the API fails.
+    """
+
+    if isinstance(protein, Protein):
+        protein = protein.uuid
+
+    workflow = stjames.PoseAnalysisMolecularDynamicsWorkflow(
+        protein_uuid=protein,
+        initial_smiles=initial_smiles,
+        num_trajectories=num_trajectories,
+        simulation_time_ns=simulation_time_ns,
+        ligand_residue_name=ligand_residue_name,
+    )
+
+    data = {
+        "name": name,
+        "folder_uuid": folder_uuid,
+        "workflow_type": "pose_analysis_md",
+        "workflow_data": workflow.model_dump(serialize_as_any=True),
+        "initial_smiles": initial_smiles,
+        "max_credits": max_credits,
+    }
+
+    with api_client() as client:
+        response = client.post("/workflow", json=data)
+        response.raise_for_status()
+        return Workflow(**response.json())
+
+
+def submit_batch_docking_workflow(
+    smiles_list: list[str],
+    protein: str | Protein,
+    pocket: list[list[float]],
+    executable: str = "qvina2",
+    scoring_function: str = "vina",
+    exhaustiveness: float = 8,
+    name: str = "Batch Docking Workflow",
+    folder_uuid: str | None = None,
+    max_credits: int | None = None,
+) -> Workflow:
+    """
+    Submits a batch docking workflow to the API.
+
+    :param smiles_list: The SMILES strings to dock.
+    :param protein: The protein to dock. Can be input as a uuid or a Protein object.
+    :param executable: Which docking implementation to use.
+    :param scoring_function: Which docking scoring function to use.
+    :param exhaustiveness: Which exhaustiveness to employ.
+    :param name: The name of the workflow.
+    :param folder_uuid: The UUID of the folder to place the workflow in.
+    :param max_credits: The maximum number of credits to use for the workflow.
+    :return: A Workflow object representing the submitted batch docking workflow.
+    :raises requests.HTTPError: if the request to the API fails.
+    """
+
+    docking_settings = {
+        "executable": executable,
+        "exhaustiveness": exhaustiveness,
+        "scoring_function": scoring_function,
+    }
+
+    workflow = stjames.BatchDockingWorkflow(
+        initial_smiles_list=smiles_list,
+        target=protein,
+        pocket=pocket,
+        docking_settings=docking_settings,
+    )
+
+    data = {
+        "name": name,
+        "folder_uuid": folder_uuid,
+        "workflow_type": "batch_docking",
+        "workflow_data": workflow.model_dump(serialize_as_any=True),
+        "max_credits": max_credits,
+    }
+
+    with api_client() as client:
+        response = client.post("/workflow", json=data)
+        response.raise_for_status()
+        return Workflow(**response.json())
+
+
+def submit_msa_workflow(
+    initial_protein_sequences: list[stjames.ProteinSequence | str],
+    output_formats: list[stjames.MSAFormat],
+    name: str = "MSA Workflow",
+    folder_uuid: str | None = None,
+    max_credits: int | None = None,
+) -> Workflow:
+    """
+    Submits a multiple sequence alignment (MSA) workflow to the API.
+
+    :param initial_protein_sequences: List of protein sequences to align, as ProteinSequence objects
+    sor strings.
+    :param output_formats: List of output formats for the resulting MSA files.
+    :param name: The name to assign to the workflow.
+    :param folder_uuid: UUID of the folder where the workflow will be stored.
+    :param max_credits: The maximum number of credits to use for the workflow.
+    :return: A Workflow object representing the submitted MSA workflow.
+    :raises HTTPError: If the API request fails.
+    """
+    workflow = stjames.MSAWorkflow(
+        initial_protein_sequences=initial_protein_sequences,
+        output_formats=output_formats,
+    )
+
+    data = {
+        "name": name,
+        "folder_uuid": folder_uuid,
+        "workflow_type": "msa",
+        "workflow_data": workflow.model_dump(serialize_as_any=True),
+        "max_credits": max_credits,
+    }
+
+    with api_client() as client:
+        response = client.post("/workflow", json=data)
+        response.raise_for_status()
+        return Workflow(**response.json())
